/* sparkl.c:  simulate pixel bloom and diffraction effects    */
/*           John P. Beale 10/31/95                           */
/*           GIF library routines from gd library             */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include "hcon.h"
#include "hf_io.h"
#include "gems.h"
#define LUMTYPE unsigned short
#define LUMMAX 65535
#define PSV 2                   /* point-spread value  */

	/* select an element of an array, clipping if out-of-bounds */

#define SetEl(vv, xq, yq, val) if ((xq >= 0) && (yq >= 0) && \
				  (xq < xsize) && (yq < ysize) ) \
				  vv[(yq*xsize) + (xq)] = val

int tga3_read(Cimg *imp, char *fname);
int tga3_write(Cimg imout, char *fname);
			
			/* generate luminosity array */
void genhsl(Cimg im1, LUMTYPE **lum);
void RGB_to_HSL (D r, D g, D b, D *h, D *s, D *l);
void HSL_to_RGB (D h, D s, D l, D *r, D *g, D *b);

					/* generate diff. effects */
int sparkl(Cimg im1,Cimg *im2p,LUMTYPE *lum,
	   D lumlim, D csatf, D lscale,D iscale);

D histc(PIXTYPE *im1, int xsize, int ysize, D frac);
D histl(LUMTYPE *lum, int xsize, int ysize, D frac); /* luminosity hist. */
void fspike(Cimg *im1p, Cimg *im2p, D xr, D yr, D intens, 
		D csatf, D rscale, D th1);

     /* draw an anti-aliased point at (real) coords xr,yr, intensity mag */
int rpnt(Cimg *im2p,D xr, D yr, D ri, D gi, D bi);

     /* draw a gaussian-profiled ring around a point */
void ring(PIXTYPE **im2p, D xr, D yr, D rad, 
	    D width, D intens);

U xsize, ysize;
PIXTYPE fmin = 0, fmax = 1.0;            /* scaling limits */
U s_points = 6;                          /* number of diffraction points */
D sp_angle = 15;                         /* initial spike angle, degrees */
Boolean norenorm = FALSE;
Boolean negate = FALSE;
Boolean rescale = FALSE;
Boolean clip = FALSE;
D pi = 3.1415926535897932385;

/* ---------------------------------- */

int main(int argc, char **argv)
{ char *usage;
  char *utemp = 
"sparkl: add diffraction effects v0.1a (c) <beale@jump.stanford.edu> 1995\n\
Usage: sparkl [ <lt> <ls> <is> <sp> <sa> ] <infile> <outfile>\n\n\
	  <Lum_Thr>   intensity threshold [0..1]    default 0.001\n\
	  <Len_Scale> spike length [pixels]         default 20\n\
	  <I_Scale>   flare intensity ratio [0..1]  default 0.5\n\
	  <Spke_Pnts> number of spike points [1...] default 6\n\
	  <Spke_Angl> initial spike angle, degrees  default 15\n\n\
  Input and output files are 24-bit uncompressed TGA format.\n\
  ----------------------------------------------------------\n\
  Example:  sparkl rough.tga shiny.tga\n\
    or,     sparkl 0.002 40 0.8 in.tga out.tga\n\
    or,     sparkl 0.001 30 0.2 4 30 in.tga out.tga\n";
  
  char fname_in[160];
  char fname_out[160];
  LUMTYPE *lum;
  Cimg im1, im2;    /* 3-plane color image (input and output) */
  int rval,i;
  D lumlim, lumper, csatf;
  Boolean lumperspec = FALSE;
  D lscale, iscale;

  usage = (char *)malloc((size_t)(1000 * sizeof(char)));
  sprintf(usage,utemp,Intype_string, Outtype_string);

  if ((argc < 3) || (argc > 8)) {
	fprintf(stderr,"%s",usage);
	exit(1);
      }

  lscale = 20;
  iscale = 0.5;
  csatf = 1.0;
  lumlim = 0.5;
  lumper = 0.001;  /* default percentile thresh = 0.1 %  */
  i = 1;  /* cmd-line index counter */
  if (argc > 3) {
      sscanf(argv[i++], "%lf", &lumper);
      lumperspec = TRUE;
      if (lumper < 0 || lumper > 1.0) {
	fprintf(stderr,"Threshold out of bounds (0..1).\n");
	exit(1);
      }
    if (argc > 4) sscanf(argv[i++], "%lf", &lscale);
    if (argc > 5) sscanf(argv[i++], "%lf", &iscale);
    if (argc > 6) sscanf(argv[i++], "%d", &s_points);
    if (argc > 7) sscanf(argv[i++], "%lf", &sp_angle);
  }
  strcpy(fname_in,argv[i++]);
  strcpy(fname_out,argv[i++]);

  printf("Using lum_per = %2.4f, length = %5.3f pixels, intens.= %5.3f\n",
     lumper, lscale, iscale);
  printf("%d diffraction points rotated %2.2f degrees.\n",s_points,sp_angle);
  
  im1.xsize = 23;  /* is the compiler happy now? I hope so. */
  im1.ysize = 25;
					/* read in heightfield */
   rval = tga3_read(&im1, fname_in);
   if (rval != 0)
     exit(rval);
   xsize = im1.xsize;
   ysize = im1.ysize;
   /* printf("returned xsize = %d ysize = %d\n",xsize,ysize); */

   /* printf("calling genhsl...\n"); */
   genhsl(im1,&lum);

   printf("%s: [%d x %d] \n",fname_in,xsize,ysize);
   lumlim = histl(lum,xsize,ysize,lumper);
   printf("%1.4f percentile intensity level = %1.3f\n",lumper,lumlim);
   
   /* printf("calling sparkl...\n"); */
   sparkl(im1,&im2,lum,lumlim,csatf,lscale,iscale);
   
   rval = tga3_write(im2, fname_out);

   return(rval);    /* return a value */

} /* end main() ----------------------------------------------- */

/* -------------------------------------------------------------------
 *  histc() -- returns that pixel value of im1[xsize,ysize] which 
 *            'frac' fraction of pixels exceed in magnitude.
 *  At the moment this routine is unused. You can throw it away. Yep.
 * -------------------------------------------------------------------
 */

D histc(PIXTYPE *im1, int xsize, int ysize, D frac)
{
 D dmax, dmin, tmp;
 D retval;
 int ix, iy;
 int hist[1001];   /* histogram with 1001 bins */
 int hlim = 1000;
 int p10;
 D p10i;
 int p10lim;
 long num_pixels;     /* total number of pixels in this image */

 for (ix=0;ix<1001;ix++) hist[ix] = 0;  /* clear hist[x]. Neccesary? */

 dmax = El(im1,0,0)/(D)PIXMAX;
 dmin = El(im1,0,0)/(D)PIXMAX;

 /* Find minimum and maximum values in array. */

   for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	tmp = (double)El(im1,ix,iy)/PIXMAX;
	hist[(int)((hlim+0.999)*tmp)]++;   /* increment this histogram bin */
	if (tmp > dmax) dmax = tmp;
	if (tmp < dmin) dmin = tmp;
    }
   }

   printf("total lum range %1.2e .. %1.2e\n",dmin,dmax);
   p10 = 0;
   num_pixels = xsize*ysize;
   p10lim = num_pixels*frac;
   ix = hlim;                            /* maximum value */
   p10i = ix;
   do {
     p10 += hist[ix]; 
     p10i = ix;
     ix--;
   } while ((p10 < p10lim) && (ix>0));

  retval = p10i/hlim;
  return(retval);

} /* end histc() */

/* -------------------------------------------------------------------
 *  histl() -- returns that pixel value of im1[xsize,ysize] which 
 *            'frac' fraction of pixels exceed in magnitude.
 * -------------------------------------------------------------------
 */

D histl(LUMTYPE *im1, int xsize, int ysize, D frac)
{
 D dmax, dmin, tmp;
 D retval;
 int ix, iy;
 int hist[1001];   /* histogram with 1001 bins */
 int hlim = 1000;
 int p10;
 D p10i;
 int p10lim;
 long num_pixels;     /* total number of pixels in this image */

 for (ix=0;ix<1001;ix++) hist[ix] = 0;  /* clear hist[x]. Neccesary? */

 dmax = El(im1,0,0)/(D)LUMMAX;
 dmin = El(im1,0,0)/(D)LUMMAX;

 /* Find minimum and maximum values in array. */

   for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	tmp = (double)El(im1,ix,iy)/LUMMAX;
	hist[(int)((hlim+0.999)*tmp)]++;   /* increment this histogram bin */
	if (tmp > dmax) dmax = tmp;
	if (tmp < dmin) dmin = tmp;
    }
   }

   printf("Full luminosity range is %1.2e .. %1.2e\n",dmin,dmax);
   p10 = 0;
   num_pixels = xsize*ysize;
   p10lim = num_pixels*frac;
   ix = hlim;                            /* maximum value */
   p10i = ix;
   do {
     p10 += hist[ix]; 
     p10i = ix;
     ix--;
   } while ((p10 < p10lim) && (ix>0));

  retval = p10i/hlim;
  return(retval);

} /* end histl() */

void genhsl(Cimg im1, LUMTYPE **lum)
{
 int ix, iy;
 size_t bufsize;
 LUMTYPE *ll;
 double r,g,b;
 double h,s,l;
 U xsize, ysize;

 xsize = im1.xsize;
 ysize = im1.ysize;
 bufsize = (size_t) (xsize * ysize * sizeof(LUMTYPE));
 ll = (LUMTYPE *) malloc(bufsize);
 if (ll == NULL) {
   fprintf(stderr,"Could not allocate %ld bytes for image.\n",(long) bufsize);
   exit(1);
  }
 lum[0] = ll;
 
 for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	r = Elr(im1,ix,iy) / (double)(PIXMAX);
	g = Elg(im1,ix,iy) / (double)(PIXMAX);
	b = Elb(im1,ix,iy) / (double)(PIXMAX);
	RGB_to_HSL(r,g,b,&h,&s,&l);
	El(ll,ix,iy) = (LUMTYPE)(l * (LUMTYPE)(LUMMAX));
    }
 }

} /* end genhsl() */

/* -------------------------------------------------------------------- */
/* sparkl() -- make new image array im2[x,y] and create processed       */
/*             copy of im1[x,y] in it.                                  */
/*             This is the main routine here. All else is, um, else.    */
/* -------------------------------------------------------------------- */
int sparkl(Cimg im1,Cimg *im2p,LUMTYPE *lum,
	   D lumlim, D csatf, D lscale,D iscale)
{
 D tmp;
 D nfrac;   /* normalized fraction */
 Cimg im2;
 int ix,iy;
 size_t bufsize;
 D length, intens;
 U xsize, ysize;

			/* create new image of same size as old image */
 xsize = im2.xsize = im1.xsize;
 ysize = im2.ysize = im1.ysize;
 bufsize = (size_t) (im1.xsize * im1.ysize * sizeof(PIXTYPE));
 im2.r = (PIXTYPE *) malloc(bufsize);
 im2.g = (PIXTYPE *) malloc(bufsize);
 im2.b = (PIXTYPE *) malloc(bufsize);
 if (im2.r == NULL || im2.g==NULL || im2.b==NULL) {
   fprintf(stderr,"Could not allocate %ld bytes for image.\n",(long) bufsize);
   exit(1);
  }
 im2p[0] = im2;
 
 printf("working..."); fflush(stdout);
   for (iy = 0; iy < ysize; iy++) {   /* copy old image to new image */
     for (ix = 0; ix < xsize; ix++) {
	 Elr(im2,ix,iy) = Elr(im1,ix,iy);  /* red, green, blue planes */
	 Elg(im2,ix,iy) = Elg(im1,ix,iy);
	 Elb(im2,ix,iy) = Elb(im1,ix,iy);
     }
   }

/* add effects to new image based on intensity of old pixels */

   for (iy = 0; iy<ysize; iy++) {  
     for (ix = 0; ix<xsize; ix++) {
	 tmp = El(lum,ix,iy)/(double)(LUMMAX);
	 if (tmp > lumlim) {
	   nfrac = ABS((tmp-lumlim)/(1-lumlim));
	   length = lscale * pow(nfrac,0.8);
	   intens = iscale * pow(nfrac,1);
	       /* fspike im x,y intens rlength angle */
	   if (s_points > 0) 
	     fspike(&im1, &im2,ix,iy,intens,csatf,length,sp_angle);

		/* ring im x,y rad width intens */
/*           ring(im2p,ix,iy,length,length/6,intens*0.7); */
	 } /* end if (tmp>lumlim) */
     }
   }
  return(0);
}       /* end sparkl */

/* ---------------------------------------------------------- 
 * ring() --  draw a smoothly-varying ring or halo
 *              !! needs to be converted from B/W to full color !!
 * im2p: pointer to image array
 * xr, yr: center coordinates
 * rad: radius of ring (pixel units)
 * width: 1/e width of ring
 * intens: intensity of ring
 * ----------------------------------------------------------
 */

void ring(PIXTYPE **im2p, D xr, D yr, D rad, 
	    D width, D intens)
{
int ix, iy;
int xylim;
D x,y;
D rdist, iloc;
PIXTYPE *im2;
D tmp;

  im2 = im2p[0];
  xylim = rad+(2*width);

   for (x = (xr-xylim); x < (xr+xylim); x++) {
     ix = (int) x;
     if (ix>=0 && ix < xsize) {
      for (y = (yr-xylim); y < (yr +xylim); y++) {
	iy = (int) y;
	if (iy>=0 && iy<ysize) {
	  rdist = sqrt((xr-x)*(xr-x) + (yr-y)*(yr-y));
	  iloc = intens * exp(-pow( (rdist-rad)/width,2)); 
	  tmp = El(im2,ix,iy)/(double)(PIXMAX);  /* increment image pixel */
	  tmp += iloc;
	  if (tmp >= 1.0) El(im2,ix,iy) = PIXMAX;  /* saturation */
		else El(im2,ix,iy) = (PIXTYPE)( tmp * PIXMAX);
	}
      } /* end for y */
     } /* end if ix */
   } /* end for x */
}



/* ---------------------------------------------------------- 
 * fspike() --  draw a diffraction spike pattern 
 * with a set of fading lines. Takes as parameters:
 * -----------------------------------------------
 * im2p: pointer to image array
 * th1: angle in degrees
 * xr,yr: start position
 * intens: pixel intensity
 * rscale: 1/e length of line (in pixels)
 * --------also uses variables:
 * s_points: number of major spike points
 * xrt,yrt: x,y temporary position
 * dx,dy: increments to xrt,yrt position variables
 * rpos:  current distance drawn
 * inloc: pixel intensity at this location
 * theta: temporary angle variable
 * efac: falloff exponent
 * ----------------------------------------------------------
 */

void fspike(Cimg *im1p, Cimg *im2p, D xr, D yr, D intens, 
		D csatf, D rscale, D th1)
{

D xrt, yrt, dx, dy, rpos, inr, ing, inb;
D valr, valg, valb;  /* floating-point red,green,blue value on [0..1] */
D ho, so, vo;        /* HSV [0..1] values for initial pixel */
D snew;              /* current value of saturation */
D theta, efac;
D sfac;              /* scaling factor 1..0 in distance along spike */
int i;
int x,y;

#define FADELINE                \
  x = (int)(xr+0.5); y=(int)(yr+0.5); \
  valr = Elr(im1p[0],x,y)/(D)PIXMAX; \
  valg = Elg(im1p[0],x,y)/(D)PIXMAX; \
  valb = Elb(im1p[0],x,y)/(D)PIXMAX; \
  RGB_to_HSL(valr,valg,valb,&ho,&so,&vo); \
  dx = 0.2* cos(theta*pi/180);  \
  dy = 0.2* sin(theta*pi/180);  \
  xrt = xr; yrt = yr;           \
  rpos = 0.2;                   \
  do {                          \
   sfac = exp(-pow(rpos/rscale,efac)); \
   snew = (sfac)*csatf + (1-sfac)*so; /* saturation interpolation */ \
   snew = 1.0; \
   sfac *= intens; \
   HSL_to_RGB(ho,snew,vo,&valr,&valg,&valb); \
   inr = 0.2 * valr * sfac; \
   ing = 0.2 * valg * sfac; \
   inb = 0.2 * valb * sfac; \
   rpnt(im2p,xrt,yrt,inr,ing,inb);              \
   xrt += dx;                                   \
   yrt += dy;                                   \
   rpos += 0.2;                                 \
  } while ((inr > 0.01) || (ing > 0.01) || (inb>0.01))              

  theta = th1;
  efac = 2.0;    /* exponent on intensity falloff with radius */

  for (i=0;i<s_points;i++) {   /* draw the major spikes */
    FADELINE;
    theta += 360.0/s_points;
  }

  theta = th1 + 180.0/s_points;  /* minor rays centered between major ones */
  rscale *= 0.7;    /* minor spikes are smaller */
  intens *= 0.7;    /* minor spikes are dimmer */

  for (i=0;i<s_points;i++) {   /* draw the minor spikes */
    FADELINE;
    theta += 360.0/s_points;
  }

} /* end fspike() */


/* ---------------------------------------------------------------
 * rpnt() -- draw a point, given real coordinates xr,yr on an integer
 * pixelmap which is specified as a pointer to im2 3-color image
 * using a gaussian point-spread function truncated to the nearest 
 * four pixels. ri,gi,bi are real-valued red,green,blue increments.
 * ---------------------------------------------------------------
 */

int rpnt(Cimg *im2p,D xr, D yr, D ri, D gi, D bi)
{
 int x,y;
 D dx,dy,rs;
 D valr, valg, valb;
 Cimg im2;

#define ADDPIXEL        dx = xr - x; dy = yr - y;  \
			rs = dx*dx + dy*dy; \
			valr = ri * exp(-rs/PSV); \
			valg = gi * exp(-rs/PSV); \
			valb = bi * exp(-rs/PSV); \
		if (x>=0 && y>=0 && x<xsize && y<ysize) { \
			valr += Elr(im2,x,y)/(double)PIXMAX; \
			valg += Elg(im2,x,y)/(double)PIXMAX; \
			valb += Elb(im2,x,y)/(double)PIXMAX; \
			if (valr > 1.0) valr = 1.0; \
			if (valg > 1.0) valg = 1.0; \
			if (valb > 1.0) valb = 1.0; \
			Elr(im2,x,y) = (PIXTYPE)(valr*PIXMAX); \
			Elg(im2,x,y) = (PIXTYPE)(valg*PIXMAX); \
			Elb(im2,x,y) = (PIXTYPE)(valb*PIXMAX); }

 im2 = im2p[0];        /* dereference the image pointer */

 x = (int) (xr);  /* integer coord. to upper left of real point */
 y = (int) (yr);

 ADDPIXEL;
 x++;                   /* upper right */
 ADDPIXEL;
 y++;                   /* lower right */
 ADDPIXEL;
 x--;                   /* lower left */
 ADDPIXEL;

 return(0);
}
